/** \file scrmode.c
  * \brief Manipulation of screen bitmap modes
  * You shouldn't use those unless you know what
  * you are doing. The non-DIRECT modes are slow
  * and don't work well. This feature may even be
  * removed in later versions of AllegroGL.
  */

#include <string.h>

#include "alleggl.h"
#include "allglint.h"
#include "glvtable.h"



/* C write/unwrite line routines -- defined in full later on */
static unsigned long agl_write_line_c(BITMAP * bmp, int line);
static void agl_unwrite_line_c(BITMAP * bmp);


/* Now if Allegro is in assembler mode, it'll call these
 * functions using a special calling convention.  We need to
 * wrap them, in that case. */
#ifdef ALLEGRO_NO_ASM
	/* Cool, we can just use our C routines directly */
       static void *agl_write_line = agl_write_line_c;
       static void *agl_unwrite_line = agl_unwrite_line_c;
#else
	/* Oh well, we need to wrap the C routines. */
#ifdef ALLEGRO_ASM_PREFIX
#define PREFIX ALLEGRO_ASM_PREFIX
#else
#define PREFIX
#endif
#if defined ALLEGRO_I386
		/* i386 is all we support at the moment */
#if defined ALLEGRO_GCC && !defined USE_PREASSEMBLED
			/* GCC syntax */
asm("\n"
	PREFIX "agl_write_line_asm:\n"
	"	pushl %ecx\n"
	"	pushl %eax\n"
	"	pushl %edx\n"
	"	call " PREFIX "agl_write_line_c\n"
	"	popl %edx\n"
	"	popl %ecx\n"	/* note: dummy pop */
	"	popl %ecx\n"
	"	ret\n"
	"\n"
	PREFIX "agl_unwrite_line_asm:\n"
	"	pushl %eax\n"
	"	pushl %ecx\n"
	"	pushl %edx\n"
	"	call " PREFIX "agl_unwrite_line_c\n"
	"	popl %edx\n"
	"	popl %ecx\n"
	"	popl %eax\n"
	"	ret\n"
);

extern void agl_write_line_asm();
extern void agl_unwrite_line_asm();
static void *agl_write_line = agl_write_line_asm;
static void *agl_unwrite_line = agl_unwrite_line_asm;
#else
			/* Unsupported syntax -- use preassembled code
			 * block.  This may or may not work. */
unsigned char preassembled_code[] = {
	0x51, 0x50, 0x52,
	0xe8, 0, 0, 0, 0,
	0x5a, 0x59, 0x59,
	0xc3,
	0x50, 0x51, 0x52,
	0xe8, 0, 0, 0, 0,
	0x5a, 0x59, 0x58,
	0xc3
};
#define PREASM
#define PREASM_OFFSET_WRITE_FUNC 4
#define PREASM_OFFSET_UNWRITE_FUNC 16
static void *agl_write_line = preassembled_code;
static void *agl_unwrite_line = preassembled_code + 12;
#endif
#else
		/* Not i386 -- assembler not supported.  Allegro
		 * will never let us get here anyway. */
static void *agl_write_line = NULL, *agl_unwrite_line = NULL;
#endif /* i386 */
#endif /* !no_asm */



static int screen_w, screen_h;

/* buffers screen image in some modes */
static BITMAP *__allegro_gl_memory = NULL;
static GFX_VTABLE memory_vtable;    /* vtable for accessing above bitmap */
static int *dirty_flags = NULL;     /* flags showing dirty lines in bitmap */
static int dirty_size = 0;          /* size (in bytes) of `dirty_flags' */


int __allegro_gl_screen_mode = AGL_MODE_DIRECT;


static int lastline = -1;

/* agl_write_line_c:
 *  This routine will be called by Allegro whenever the user
 *  tries to write to a line of the `__allegro_gl_memory' bitmap.
 */
static unsigned long agl_write_line_c(BITMAP * bmp, int line)
{
	switch (__allegro_gl_screen_mode) {

		case AGL_MODE_DIRECT:
			break;

		case AGL_MODE_DUPLEX:
			/* copy the line from the framebuffer using OpenGL */
			if (lastline != -1)
				agl_unwrite_line_c(bmp);
			lastline = line;

			glReadPixels(0, screen_h - lastline, screen_w, 1,
			             __allegro_gl_get_bitmap_color_format(screen, 0),
			             __allegro_gl_get_bitmap_type(screen, 0),
			             bmp->line[lastline]);

			break;

		case AGL_MODE_OPAQUE:
		case AGL_MODE_MASKED:
			/* mark the line as dirty */
			dirty_flags[line] = 1;
			break;

		case AGL_MODE_OPAQUE_RT:
		case AGL_MODE_MASKED_RT:
			if (lastline != -1)
				agl_unwrite_line_c(bmp);
			lastline = line;
			break;

	}
	return (int) bmp->line[line];
}



/* agl_unwrite_line_c:
 *  This is called by Allegro when the user has finished with a
 *  line of the `__allegro_gl_memory' bitmap.
 */
static void agl_unwrite_line_c(BITMAP * bmp)
{
	switch (__allegro_gl_screen_mode) {

		case AGL_MODE_DIRECT:
			break;

		case AGL_MODE_DUPLEX:
		case AGL_MODE_OPAQUE_RT:
			/* paste the line back to the framebuffer using OpenGL */
			if (lastline != -1) {
				allegro_gl_begin();
				glRasterPos2i(0, lastline);
				glDrawPixels(screen_w, 1,
				    __allegro_gl_get_bitmap_color_format(screen, 0),
				    __allegro_gl_get_bitmap_type(screen, 0),
				    bmp->line[lastline]);
				allegro_gl_end();
				lastline = -1;
			}
			break;

		case AGL_MODE_MASKED_RT:
			/* Should be like above, but with a masked draw */
			break;

		case AGL_MODE_OPAQUE:
		case AGL_MODE_MASKED:
			break;

	}
}



/* prepare_for_blitting:
 *  This routine is meant to streamline everything for calls to 
 *  glDrawPixels.
 */
static void prepare_for_blitting(void)
{
	allegro_gl_begin();

//      glPushAttrib (GL_ENABLE_BIT | GL_PIXEL_MODE_BIT);

	glDisable(GL_ALPHA_TEST);
	glDisable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_FOG);
	glDisable(GL_LIGHTING);
#ifdef GL_COLOR_LOGIC_OP
	glDisable(GL_COLOR_LOGIC_OP);
#endif
	glDisable(GL_STENCIL_TEST);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);

	glPixelTransferi(GL_MAP_COLOR, GL_FALSE);
	glPixelTransferi(GL_RED_SCALE, 1);
	glPixelTransferi(GL_RED_BIAS, 0);
	glPixelTransferi(GL_GREEN_SCALE, 1);
	glPixelTransferi(GL_GREEN_BIAS, 0);
	glPixelTransferi(GL_BLUE_SCALE, 1);
	glPixelTransferi(GL_BLUE_BIAS, 0);
	glPixelTransferi(GL_ALPHA_SCALE, 1);
	glPixelTransferi(GL_ALPHA_BIAS, 0);

	glPixelZoom(1.0, 1.0);

	allegro_gl_end();
}



/* init_screen_mode:
 *  Initialises the given update mode.  0 = success.
 */
static int init_screen_mode(int newmode)
{
	switch (newmode) {

		case AGL_MODE_DIRECT:
			screen = allegro_gl_screen;
			return 0;

		case AGL_MODE_DUPLEX:
			screen = __allegro_gl_memory;
			prepare_for_blitting();
			return 0;

		case AGL_MODE_OPAQUE:
		case AGL_MODE_OPAQUE_RT:
			screen = __allegro_gl_memory;
			clear_bitmap(screen);
			memset(dirty_flags, 0, dirty_size);
			return 0;

		case AGL_MODE_MASKED:
		case AGL_MODE_MASKED_RT:
			screen = __allegro_gl_memory;
			clear_to_color(screen, bitmap_mask_color(screen));
			memset(dirty_flags, 0, dirty_size);
			return 0;

		default:
			return -1;

	}
}



/* shutdown_screen_mode:
 *  Cleans up, if necessary, after the given update mode.  0 = success.
 */
static int shutdown_screen_mode(int oldmode)
{
	switch (oldmode) {

		case AGL_MODE_DIRECT:
			return 0;

		case AGL_MODE_DUPLEX:
		case AGL_MODE_OPAQUE_RT:
		case AGL_MODE_MASKED_RT:
			if (lastline != -1)
				agl_unwrite_line_c(screen);
			return 0;

		case AGL_MODE_OPAQUE:
			glRasterPos2i(0, 0);
			glDrawPixels(screen_w, screen_h,
			    __allegro_gl_get_bitmap_color_format(screen, 0),
			    __allegro_gl_get_bitmap_type(screen, 0), screen->line[0]);
			return 0;

		case AGL_MODE_MASKED:
			/* as above but masked */
			return 0;

		default:
			return -1;

	}
}



/** \defgroup updatemodes Screen update modes
  *
  * True OpenGL programs don't need to use these, and none of them work
  * very well.  My recommendation is simply to use OpenGL graphics calls to
  * write to the screen -- that's why you're using AllegroGL after all, isn't
  * it?
  *
  *  AllegroGL provides the global `allegro_gl_screen', which
  *  always points to what is initially returned as the screen
  *  bitmap -- you can draw onto this with (some) standard
  *  Allegro calls, and they get converted directly into OpenGL
  *  calls.  I call this `direct' mode.
  *
  *  The other options revolve around `__allegro_gl_memory', a
  *  memory bitmap which is drawn to like any other memory
  *  bitmap.  The exact behaviour here is determined by the
  *  precise mode selected, and this is all very experimental at
  *  the moment.
  *
  * \{ */

/** \def AGL_MODE_DIRECT
  * \brief Direct mode
  * 
  * In this mode, <em>screen</em> just points at allegro_gl_screen.  Allegro
  * graphics calls acting on it are translated to similar
  * OpenGL calls using a special (incomplete) vtable.  If you call functions
  * using unimplemented vtable entries, your program will crash.
  *
  * \deprecated
  */
/** \def AGL_MODE_DUPLEX
  * \brief Duplex mode
  * 
  * Allegro graphics calls acting on <em>screen</em> are performed in a memory
  * buffer, which is filled from the OpenGL framebuffer before the call and
  * pasted back afterwards.  This is rather slow.
  *
  * \deprecated
  */
/** \def AGL_MODE_OPAQUE
  * \brief Opaque mode
  * 
  * As for Duplex mode, but the buffer is not filled from the OpenGL
  * framebuffer, nor is it copied back there on the fly -- it is
  * copied only when the update mode is changed again.  The drawing is 
  * opaque, so any OpenGL-drawn graphics will be cleared.
  *
  * \deprecated
  */
/** \def AGL_MODE_MASKED
  * \brief Masked mode
  * 
  * As for Opaque mode, but the copy is performed in a masked drawing mode,
  * so transparent pixels in your Allegro graphics will not obscure the
  * OpenGL framebuffer.  I doubt this actually works though.
  *
  * \deprecated
  */
/** \def AGL_MODE_OPAQUE_RT
  * \brief Opaque realtime mode
  * 
  * Similar to Opaque mode, but your graphics are copied to the OpenGL
  * framebuffer as soon as you finish writing to each line.  In this way
  * it is a `half-duplex' mode.  This is probably not implemented.
  *
  * \deprecated
  */
/** \def AGL_MODE_MASKED_RT
  * \brief Masked realtime mode
  * 
  * Similar to Masked mode, but your graphics are copied to the OpenGL 
  * framebuffer as soon as you finish writing to each line.  AFAICS this
  * concept is fundamentally flawed, and it is probably not implemented.
  *
  * \deprecated
  */

/** \} */


/** \ingroup updatemodes
  *
  * \brief Sets the screen update mode
  *
  * \return int Zero on success, nonzero on failure
  *
  * \deprecated This function will be removed in a later version of AllegroGL
  */
int allegro_gl_screen_mode(int newmode)
{
	return shutdown_screen_mode(__allegro_gl_screen_mode)
		|| init_screen_mode(__allegro_gl_screen_mode = newmode);
}



/* create_memory_screen:
 *  Creates the memory screen bitmap and sets up the write/unwrite line
 *  functions.
 */
static void create_memory_screen(void)
{
	__allegro_gl_memory = create_bitmap(screen_w, screen_h);
	if (!__allegro_gl_memory)
		return;

	clear_bitmap(__allegro_gl_memory);

	dirty_size = screen_h * sizeof *dirty_flags;
	dirty_flags = malloc(dirty_size);
	if (!dirty_flags) {
		destroy_bitmap(__allegro_gl_memory);
		__allegro_gl_memory = NULL;
		return;
	}

	{
		int i;
		for (i = 0; i < screen_h; i++)
			dirty_flags[i] = 0;
	}

	memcpy(&memory_vtable, __allegro_gl_memory->vtable, sizeof memory_vtable);
	__allegro_gl_memory->vtable = &memory_vtable;
	__allegro_gl_memory->write_bank = agl_write_line;
	__allegro_gl_memory->vtable->unwrite_bank = agl_unwrite_line;
}



/* __allegro_gl_init_screen_mode:
 *  Initialise this module.
 */
void __allegro_gl_init_screen_mode(void)
{
	static int initialised = 0;
	if (initialised) {
		/* Shut down before reinitialising */
		if (__allegro_gl_memory) {
			destroy_bitmap(__allegro_gl_memory);
			__allegro_gl_memory = NULL;
			free(dirty_flags);
			dirty_flags = NULL;
		}
	}
	else {
		/* One-off initialisation */
		initialised = 1;

#ifdef PREASM
		/* In the preassembled code, we can't use the linker
		 * directly to fill in the addresses of the C routines,
		 * so we have to do it indirectly like this. */
		*(int *) (preassembled_code + PREASM_OFFSET_WRITE_FUNC) =
			(int) agl_write_line_c - (int) preassembled_code -
			PREASM_OFFSET_WRITE_FUNC - 4;
		*(int *) (preassembled_code + PREASM_OFFSET_UNWRITE_FUNC) =
			(int) agl_unwrite_line_c - (int) preassembled_code -
			PREASM_OFFSET_UNWRITE_FUNC - 4;
#else
		/* Suppress warnings. */
		(void) agl_write_line_c;
		(void) agl_unwrite_line_c;
#endif
	}

	screen_w = allegro_gl_display_info.w;
	screen_h = allegro_gl_display_info.h;
	create_memory_screen();

	init_screen_mode(AGL_MODE_DIRECT);
}



/* __allegro_gl_release_screen:
 * Check the screen update mode and free whichever of __allegro_gl_memory
 * and allegro_gl_screen isn't active.
 */
 void __allegro_gl_release_screen(void)
 {
	switch(__allegro_gl_screen_mode) {

		case AGL_MODE_DIRECT:
			/* screen == allegro_gl_screen */
			if (__allegro_gl_memory)
				destroy_bitmap(__allegro_gl_memory);
			break;

		case AGL_MODE_DUPLEX:
		case AGL_MODE_OPAQUE_RT:
		case AGL_MODE_MASKED_RT:
		case AGL_MODE_OPAQUE:
		case AGL_MODE_MASKED:
			/* screen == __allegro_gl_memory */
			if (allegro_gl_screen)
				destroy_bitmap(allegro_gl_screen);
			break;

		default:
			return;
			
	}
	
	if (dirty_flags)
		free(dirty_flags);
	dirty_flags = NULL;

	/* Allegro will destroy the screen bitmap,
	   so the pointers are now useless */
	__allegro_gl_memory = NULL;
	allegro_gl_screen = NULL;
 }

