/** \file videovtb.c
  * \brief video bitmaps (ie. texture rendering) vtable
  *  Some of these don't work correctly or will be very slow.
  *  The current vtable is in early stages, and will be replaced by a better
  *  (faster) version 'soon'. Don't count on regular Allegro functions
  *  working on the screen or on video bitmaps.
  */

#include <string.h>

#include <allegro.h>

#ifdef ALLEGRO_WINDOWS
#include <winalleg.h>
#endif

#include "alleggl.h"
#include "allglint.h"
#include "glvtable.h"
#include <allegro/internal/aintern.h>
#ifdef ALLEGRO_MACOSX
#include <OpenGL/glu.h>
#else
#include <GL/glu.h>
#endif


static GFX_VTABLE allegro_gl_video_vtable;

/* Counter for video bitmaps. screen = 1 */
static int video_bitmap_count = 2;


/* Will make a piece-wise texture out of the specified bitmap
 * Source -must- be a memory bitmap or memory subbitmap (created by Allegro
 * only).
 *
 * And must be a 32bpp image
 */
static BITMAP *allegro_gl_make_video_bitmap(BITMAP *bmp) {
	
	int i, j;

	/* XXX <rohannessian> If ARB_np2 is support, just use that.
	 * Is [EXT|NV]_texture_rectangle is available, use that.
	 * 
	 * Otherwise, try this algo instead:
	 *  1. Round to next higher power of 2, if not power-of-2
	 *  2. If wasted space > 25% of area
	 *   2.1 Split into 4 textures using lower nearest power-of-2
	 *   2.2 Remainder is rounded to next higher power-of-2
	 *   2.3 Repeat step 2.
	 *
	 *  We also need to check for the implementation's highest texture size.
	 */
	int n_w = (bmp->w + 255) / 256,
	    n_h = (bmp->h + 255) / 256;
		
	AGL_VIDEO_BITMAP **pvid = (AGL_VIDEO_BITMAP**)&bmp->extra;
	
	GLuint old_bind;
	glGetIntegerv(GL_TEXTURE_BINDING_2D, &old_bind);
	
	for (j = 0; j < n_h; j++) {
		for (i = 0; i < n_w; i++) {
			BITMAP *mem;
			
			(*pvid) = malloc(sizeof(AGL_VIDEO_BITMAP));
			
			if (!(*pvid))
				goto agl_error;
				
			memset(*pvid, 0, sizeof(AGL_VIDEO_BITMAP));
			
			/* Create associated bitmap */
			(*pvid)->memory_copy = create_bitmap_ex(32, 
				((i < (n_w - 1)) || ((bmp->w & 255) == 0))
				    ? 256 : __allegro_gl_make_power_of_2(bmp->w & 255),
				((j < (n_h - 1)) || ((bmp->h & 255) == 0))
				    ? 256 : __allegro_gl_make_power_of_2(bmp->h & 255));
				
			if (!(*pvid)->memory_copy)
				goto agl_error;
				
			/* Fill in some values in the bitmap to make it act as a subbitmap
			 */
			mem = (*pvid)->memory_copy;
			(*pvid)->x_ofs = i * 256;
			(*pvid)->y_ofs = j * 256;

			/* Make a texture out of it */
			glGenTextures(1, &((*pvid)->tex));
			if (!((*pvid)->tex))
				goto agl_error;

			glBindTexture(GL_TEXTURE_2D, ((*pvid)->tex));

			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, mem->w, mem->h,
					0, GL_RGBA,	GL_UNSIGNED_BYTE, mem->line[0]);					

			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
					
			pvid = &((*pvid)->next);
		}
	}
	glBindTexture(GL_TEXTURE_2D, old_bind);
	(*pvid) = NULL;
	
	return bmp;
	
	agl_error: {
		AGL_VIDEO_BITMAP *vid = bmp ? bmp->extra : NULL, *next;

		while (vid) {
			if (vid->memory_copy)
				destroy_bitmap(vid->memory_copy);
				
			if (vid->tex)
				glDeleteTextures(1, &vid->tex);
								
			next = vid->next;
			free(vid);
			vid = next;
		}
	}
	return NULL;
}



/* BITMAP *allegro_gl_create_video_bitmap(int w, int h) */
/** create_video_bitmap() overload.
  * This function will create a video bitmap using texture memory.
  * Video bitmaps do not currently share space with the frame buffer (screen).
  * Video bitmaps may be of any size, but the color depth is fixed at 32bpp no
  * matter what the frame buffer's depth might be.
  * Video bitmaps are lost when switching screen modes.
  */
BITMAP *allegro_gl_create_video_bitmap(int w, int h) {

	BITMAP *bitmap;
	
	bitmap = malloc(sizeof(BITMAP) + sizeof(char *));
	
	if (!bitmap)
		return NULL;

	bitmap->dat = NULL;
	bitmap->w = bitmap->cr = w;
	bitmap->h = bitmap->cb = h;
	bitmap->clip = TRUE;
	bitmap->cl = bitmap->ct = 0;
	bitmap->write_bank = bitmap->read_bank = NULL;
	/* We should keep tracks of allocated bitmaps for the ref counter */
	bitmap->id = BMP_ID_VIDEO | video_bitmap_count;
	bitmap->extra = NULL;
	bitmap->x_ofs = 0;
	bitmap->y_ofs = 0;
	bitmap->seg = _default_ds();
	bitmap->line[0] = NULL;

	if (!allegro_gl_make_video_bitmap(bitmap)) {
		free(bitmap);
		return NULL;
	}
	video_bitmap_count++;
	
	/* XXX <rohannessian> We ought to leave the Allegro values intact
	 * Avoids bad interaction with correct Allegro programs.
	 */
	allegro_gl_video_vtable.color_depth = 32;
	allegro_gl_video_vtable.mask_color = makecol32(255, 0, 255);
	bitmap->vtable = &allegro_gl_video_vtable;

	return bitmap;
}



/* void allegro_gl_destroy_video_bitmap(BITMAP *bmp) */
/** destroy_video_bitmap() overload.
  * Will destroy the video bitmap. You mustn't use the BITMAP pointer after
  * calling this function.
  */
void allegro_gl_destroy_video_bitmap(BITMAP *bmp) {

	AGL_VIDEO_BITMAP *vid = bmp ? bmp->extra : NULL, *next;
	
	if (!bmp)
		return;
	
	while (vid) {
		if (vid->memory_copy)
			destroy_bitmap(vid->memory_copy);
				
		if (vid->tex)
			glDeleteTextures(1, &vid->tex);
								
		next = vid->next;
		free(vid);
		vid = next;
	}
	
	free(bmp);
	
	return;
}



/* static void allegro_gl_video_acquire(struct BITMAP *bmp) */
/** \ingroup glvtable
  * acquire_bitmap(bmp) overload.
  * This doesn't do anything, since OpenGL textures
  * doesn't need locking. You don't need to call this function
  * in your program.
  */
static void allegro_gl_video_acquire(struct BITMAP *bmp) {}



/* static void allegro_gl_video_release(struct BITMAP *bmp) */
/** \ingroup glvtable
  * release_bitmap(bmp) overload.
  * This doesn't do anything, since OpenGL textures
  * doesn't need locking. You don't need to call this function
  * in your program.
  */
static void allegro_gl_video_release(struct BITMAP *bmp) {}



static int allegro_gl_video_getpixel(struct BITMAP *bmp, int x, int y)
{
	int pix = -1;
	AGL_VIDEO_BITMAP *vid;
	AGL_LOG(2, "glvtable.c:allegro_gl_screen_getpixel\n");	
	
	vid = bmp->extra;
	
	/* TODO: support for subbitmaps and clipping */
	
	while (vid) {
		if (vid->x_ofs <= x && vid->y_ofs <= y
			&& vid->x_ofs + vid->memory_copy->w > x
			&& vid->y_ofs + vid->memory_copy->h > y) {
			
			pix = getpixel(vid->memory_copy, x - vid->x_ofs, y - vid->y_ofs);
			break;
		}
		vid = vid->next;
	}
	
	if (pix != -1) {
		return makeacol_depth(bitmap_color_depth(screen),
		                      getr_depth(32, pix), getg_depth(32, pix),
		                      getb_depth(32, pix), geta_depth(32, pix));
	}
	
	return -1;
}



static void allegro_gl_video_putpixel(struct BITMAP *bmp, int x, int y,
                                      int color) {
	GLbyte pixel[4];
	AGL_VIDEO_BITMAP *vid;

	split_color(color, &pixel[0], &pixel[1], &pixel[2], &pixel[3],
	            bitmap_color_depth(screen));
	vid = bmp->extra;
	
	/* TODO: support for subbitmaps and clipping */
	
	while (vid) {
		if (vid->x_ofs <= x && vid->y_ofs <= y
			&& vid->x_ofs + vid->memory_copy->w > x
			&& vid->y_ofs + vid->memory_copy->h > y) {
			
			putpixel(vid->memory_copy, x, y,
			        makeacol_depth(32, pixel[0], pixel[1], pixel[2], pixel[3]));
			
			glBindTexture(GL_TEXTURE_2D, vid->tex);
			glTexSubImage2D(GL_TEXTURE_2D, 0, x - vid->x_ofs,
				y - vid->y_ofs, 1, 1, GL_RGBA, GL_UNSIGNED_BYTE, &pixel);
				
			break;
		}
		vid = vid->next;
	}
	
	glBindTexture(GL_TEXTURE_2D, 0);
	
	return;
}



static void allegro_gl_video_vline(BITMAP *bmp, int x, int y1, int y2,
                                   int color) {

	GLubyte pixel[4];
	AGL_VIDEO_BITMAP *vid;
	GLint saved_row_length;
	
	AGL_LOG(2, "glvtable.c:allegro_gl_video_vline\n");
	split_color(color, &pixel[0], &pixel[1], &pixel[2], &pixel[3],
	            bitmap_color_depth(screen));
	vid = bmp->extra;
	
	glGetIntegerv(GL_UNPACK_ROW_LENGTH, &saved_row_length);
	
	/* TODO: support for subbitmaps and clipping */
	
	if (y1 > y2) {
		int temp = y1;
		y1 = y2;
		y2 = temp;
	}
	
	while (vid) {
		int _y1, _y2, _x;
		if (vid->x_ofs > x || vid->y_ofs > y2
			|| vid->x_ofs + vid->memory_copy->w <= x
			|| vid->y_ofs + vid->memory_copy->h <= y1) {
			
			vid = vid->next;
			continue;
		}
		
		_y1 = MAX(y1, vid->y_ofs) - vid->y_ofs;
		_y2 = MIN(y2, vid->y_ofs + vid->memory_copy->h - 1)	- vid->y_ofs;
		_x = x - vid->x_ofs;

		vline(vid->memory_copy, _x, _y1, _y2,
			makeacol_depth(32, pixel[0], pixel[1], pixel[2], pixel[3]));
		
		/* XXX <rohannessian> We should remove the hard-coded numbers here */
		glPixelStorei(GL_UNPACK_ROW_LENGTH,
		           (vid->memory_copy->line[1] - vid->memory_copy->line[0]) / 4);
			
		glBindTexture(GL_TEXTURE_2D, vid->tex);
		glTexSubImage2D(GL_TEXTURE_2D, 0, _x, _y1, 1, _y2 - _y1 + 1, GL_RGBA,
			GL_UNSIGNED_BYTE, vid->memory_copy->line[_y1] + _x * 4);

		vid = vid->next;
	}
	
	glPixelStorei(GL_UNPACK_ROW_LENGTH,	saved_row_length);
	glBindTexture(GL_TEXTURE_2D, 0);
	
	return;
}



static void allegro_gl_video_hline(BITMAP *bmp, int x1, int y, int x2,
                                   int color) {

	GLubyte pixel[4];
	AGL_VIDEO_BITMAP *vid;
	
	AGL_LOG(2, "glvtable.c:allegro_gl_video_hline\n");
	split_color(color, &pixel[0], &pixel[1], &pixel[2], &pixel[3],
	            bitmap_color_depth(screen));
	vid = bmp->extra;
	
	/* TODO: support for subbitmaps and clipping */
	
	if (x1 > x2) {
		int temp = x1;
		x1 = x2;
		x2 = temp;
	}
	
	while (vid) {
		int _x1, _x2, _y;
		if (vid->y_ofs > y || vid->x_ofs > x2
			|| vid->x_ofs + vid->memory_copy->w <= x1
			|| vid->y_ofs + vid->memory_copy->h <= y) {
			
			vid = vid->next;
			continue;
		}
		
		_x1 = MAX(x1, vid->x_ofs) - vid->x_ofs;
		_x2 = MIN(x2, vid->x_ofs + vid->memory_copy->w - 1)	- vid->x_ofs;
		_y = y - vid->y_ofs;

		hline(vid->memory_copy, _x1, _y, _x2,
			makeacol_depth(32, pixel[0], pixel[1], pixel[2], pixel[3]));
		
		
		glBindTexture(GL_TEXTURE_2D, vid->tex);
		glTexSubImage2D(GL_TEXTURE_2D, 0, _x1, _y, _x2 - _x1 + 1, 1, GL_RGBA,
			GL_UNSIGNED_BYTE, vid->memory_copy->line[_y] + _x1 * 4);

		vid = vid->next;
	}
	
	glBindTexture(GL_TEXTURE_2D, 0);
	
	return;
}



static void allegro_gl_video_line(struct BITMAP *bmp, int x1, int y1, int x2,
                                  int y2, int color) {
	
	/* Note: very very slow */						
	do_line(bmp, x1, y1, x2, y2, color, allegro_gl_video_putpixel);
	
	return;
}
	


static void allegro_gl_video_rectfill(struct BITMAP *bmp, int x1, int y1,
                                      int x2, int y2, int color) {

	GLubyte pixel[4];
	AGL_VIDEO_BITMAP *vid;
	GLint saved_row_length;
	
	AGL_LOG(2, "glvtable.c:allegro_gl_video_rectfill\n");
	split_color(color, &pixel[0], &pixel[1], &pixel[2], &pixel[3],
	            bitmap_color_depth(screen));
	vid = bmp->extra;
	
	glGetIntegerv(GL_UNPACK_ROW_LENGTH, &saved_row_length);
	
	/* TODO: support for subbitmaps and clipping */
	
	if (y1 > y2) {
		int temp = y1;
		y1 = y2;
		y2 = temp;
	}

	if (x1 > x2) {
		int temp = x1;
		x1 = x2;
		x2 = temp;
	}
	
	while (vid) {
		int _y1, _y2, _x1, _x2;
		if (vid->x_ofs > x2 || vid->y_ofs > y2
			|| vid->x_ofs + vid->memory_copy->w <= x1
			|| vid->y_ofs + vid->memory_copy->h <= y1) {
			
			vid = vid->next;
			continue;
		}
		
		_y1 = MAX(y1, vid->y_ofs) - vid->y_ofs;
		_y2 = MIN(y2, vid->y_ofs + vid->memory_copy->h - 1)	- vid->y_ofs;
		_x1 = MAX(x1, vid->x_ofs) - vid->x_ofs;
		_x2 = MIN(x2, vid->x_ofs + vid->memory_copy->w - 1)	- vid->x_ofs;

		rectfill(vid->memory_copy, _x1, _y1, _x2, _y2,
			makeacol_depth(32, pixel[0], pixel[1], pixel[2], pixel[3]));
		
		glPixelStorei(GL_UNPACK_ROW_LENGTH,
		           (vid->memory_copy->line[1] - vid->memory_copy->line[0]) / 4);
			
		glBindTexture(GL_TEXTURE_2D, vid->tex);
		glTexSubImage2D(GL_TEXTURE_2D, 0,
		    _x1, _y1, _x2 - _x1 + 1, _y2 - _y1 + 1, GL_RGBA,
			GL_UNSIGNED_BYTE, vid->memory_copy->line[_y1] + _x1 * 4);

		vid = vid->next;
	}
	
	glPixelStorei(GL_UNPACK_ROW_LENGTH,	saved_row_length);
	glBindTexture(GL_TEXTURE_2D, 0);
	
	return;
}



static int allegro_gl_video_triangle(struct BITMAP *bmp, int x1, int y1,
                                     int x2, int y2, int x3, int y3, int color)
{	
	GLubyte pixel[4];
	AGL_VIDEO_BITMAP *vid;
	GLint saved_row_length;
	int min_y, max_y, min_x, max_x;
	
	AGL_LOG(2, "glvtable.c:allegro_gl_video_triangle\n");
	split_color(color, &pixel[0], &pixel[1], &pixel[2], &pixel[3],
	            bitmap_color_depth(screen));
	vid = bmp->extra;
	
	glGetIntegerv(GL_UNPACK_ROW_LENGTH, &saved_row_length);

	min_y = MIN(y1, MIN(y2, y3));
	min_x = MIN(x1, MIN(x2, x3));
	max_y = MAX(y1, MAX(y2, y3));
	max_x = MAX(x1, MAX(x2, x3));
	
	/* TODO: Support for subbitmaps and clipping */
	
	while (vid) {
		int _y1, _y2, _x1, _x2, _x3, _y3;
		if (vid->x_ofs > max_x || vid->y_ofs > max_y
			|| vid->x_ofs + vid->memory_copy->w <= min_x
			|| vid->y_ofs + vid->memory_copy->h <= min_y) {
			
			vid = vid->next;
			continue;
		}				
		
		_y1 = y1 - vid->y_ofs;
		_y2 = y2 - vid->y_ofs;
		_y3 = y3 - vid->y_ofs;
		_x1 = x1 - vid->x_ofs;
		_x2 = x2 - vid->x_ofs;
		_x3 = x3 - vid->x_ofs;

		triangle(vid->memory_copy, _x1, _y1, _x2, _y2, _x3, _y3,
			makeacol_depth(32, pixel[0], pixel[1], pixel[2], pixel[3]));
		
		glPixelStorei(GL_UNPACK_ROW_LENGTH,
		           (vid->memory_copy->line[1] - vid->memory_copy->line[0]) / 4);
		
		/* Not quite the minimal rectangle occupied by the triangle, but
		 * pretty close
		 */
		_y1 = MAX(0, min_y - vid->y_ofs);
		_y2 = MIN(vid->memory_copy->h, max_y - vid->y_ofs);
		_x1 = MAX(0, min_x - vid->x_ofs);
		_x2 = MIN(vid->memory_copy->w, max_x - vid->x_ofs);
			
		glBindTexture(GL_TEXTURE_2D, vid->tex);
		glTexSubImage2D(GL_TEXTURE_2D, 0,
		    _x1, _y1, _x2 - _x1 + 1, _y2 - _y1 + 1, GL_RGBA,
			GL_UNSIGNED_BYTE, vid->memory_copy->line[_y1] + _x1 * 4);

		vid = vid->next;
	}
	
	glPixelStorei(GL_UNPACK_ROW_LENGTH,	saved_row_length);
	glBindTexture(GL_TEXTURE_2D, 0);	
	
	return 1;
}



void allegro_gl_video_blit_from_memory(struct BITMAP *source,
                struct BITMAP *dest, int source_x, int source_y,
				int dest_x, int dest_y, int width, int height) {

	AGL_VIDEO_BITMAP *vid;
	BITMAP *dest_parent = dest;
	GLint saved_row_length;
	
	AGL_LOG(2, "glvtable.c:allegro_gl_video_blit_from_memory\n");

	if (is_sub_bitmap (dest)) {
	   dest_x += dest->x_ofs;
	   dest_y += dest->y_ofs;
	   while (dest_parent->id & BMP_ID_SUB)
	      dest_parent = (BITMAP *)dest_parent->extra;
	}
	vid = dest_parent->extra;
	
	glGetIntegerv(GL_UNPACK_ROW_LENGTH, &saved_row_length);
	
	/* TODO: support for clipping */

	while (vid) {
		int _x, _y, _w, _h;
		if (vid->x_ofs >= dest_x + width || vid->y_ofs >= dest_y + height
			|| vid->x_ofs + vid->memory_copy->w <= dest_x
			|| vid->y_ofs + vid->memory_copy->h <= dest_y) {
			
			vid = vid->next;
			continue;
		}

		_x = MAX (vid->x_ofs, dest_x) - vid->x_ofs;
		_w = MIN (vid->x_ofs + vid->memory_copy->w, dest_x + width)
		   - vid->x_ofs - _x;
		_y = MAX (vid->y_ofs, dest_y) - vid->y_ofs;
		_h = MIN (vid->y_ofs + vid->memory_copy->h, dest_y + height)
		   - vid->y_ofs - _y;

		blit(source, vid->memory_copy, source_x + vid->x_ofs + _x - dest_x,
			source_y + vid->y_ofs + _y - dest_y, _x, _y, _w, _h);

		glPixelStorei(GL_UNPACK_ROW_LENGTH,
		           (vid->memory_copy->line[1] - vid->memory_copy->line[0]) / 4);

		glBindTexture(GL_TEXTURE_2D, vid->tex);
		glTexSubImage2D(GL_TEXTURE_2D, 0, _x, _y, _w, _h,
			GL_RGBA, GL_UNSIGNED_BYTE,
			vid->memory_copy->line[_y] + _x * 4);
			
		vid = vid->next;
	}
	
	glPixelStorei(GL_UNPACK_ROW_LENGTH,	saved_row_length);
	glBindTexture(GL_TEXTURE_2D, 0);

	return;	
}



void allegro_gl_video_blit_to_memory(struct BITMAP *source, struct BITMAP *dest,
                         int source_x, int source_y, int dest_x, int dest_y,
                         int width, int height) {

	AGL_VIDEO_BITMAP *vid;
	
	int i, j;	
	int n_w = (source->w + 255) / 256,
		n_h = (source->h + 255) / 256;
	
	AGL_LOG(2, "glvtable.c:allegro_gl_video_blit_to_memory\n");
	vid = source->extra;
	
	/* TODO: support for subbitmaps and clipping */
	
	for (j = 0; j < n_h; j++) {
		for (i = 0; i < n_w; i++) {
		
			blit(vid->memory_copy, dest, i * 256 > source_x ? 0 : source_x
					- i * 256, j * 256 > source_y ? 0 : source_y - j * 256,
				dest_x + (i == 0 ? 0 : i * 256 - source_x),
				dest_y + (j == 0 ? 0 : j * 256 - source_y),
				width - i * 256, height - j * 256);
			
			vid = vid->next;
		}
	}

	return;	
}



static void allegro_gl_video_clear_to_color(BITMAP *bmp, int color) {

	GLubyte pixel[4];
	AGL_VIDEO_BITMAP *vid;
	GLint saved_row_length;
	
	glGetIntegerv(GL_UNPACK_ROW_LENGTH, &saved_row_length);
	
	AGL_LOG(2, "glvtable.c:allegro_gl_video_clear_to_color\n");
	split_color(color, &pixel[0], &pixel[1], &pixel[2], &pixel[3],
	            bitmap_color_depth(screen));
	vid = bmp->extra;
	
	/* TODO: support for video sub bitmaps and clipping */
	
	while (vid) {
		clear_to_color(vid->memory_copy,
		            makeacol_depth(32, pixel[0], pixel[1], pixel[2], pixel[3]));
		
		glBindTexture(GL_TEXTURE_2D, vid->tex);
		
		glPixelStorei(GL_UNPACK_ROW_LENGTH,
		           (vid->memory_copy->line[1] - vid->memory_copy->line[0]) / 4);
		
		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0,
		    vid->memory_copy->w, vid->memory_copy->h,
			GL_RGBA, GL_UNSIGNED_BYTE, vid->memory_copy->line[0]);

		vid = vid->next;
	}
	
	glBindTexture(GL_TEXTURE_2D, 0);
	glPixelStorei(GL_UNPACK_ROW_LENGTH,	saved_row_length);

	return;
}



static void dummy_unwrite_bank(void)
{
}



static GFX_VTABLE allegro_gl_video_vtable = {
	0,
	0,
	dummy_unwrite_bank,			//void *unwrite_bank;  /* C function on some machines, asm on i386 */
	NULL,						//AL_METHOD(void, set_clip, (struct BITMAP *bmp));
	allegro_gl_video_acquire,
	allegro_gl_video_release,
	NULL,						//AL_METHOD(struct BITMAP *, create_sub_bitmap, (struct BITMAP *parent, int x, int y, int width, int height));
	allegro_gl_created_sub_bitmap,
	allegro_gl_video_getpixel,
	allegro_gl_video_putpixel,
	allegro_gl_video_vline,
	allegro_gl_video_hline,
	allegro_gl_video_hline,
	allegro_gl_video_line,
#if GET_ALLEGRO_VERSION() >= MAKE_VER(4, 1, 13)
	allegro_gl_video_line,
#endif	
	allegro_gl_video_rectfill,
	allegro_gl_video_triangle,
	NULL,/*allegro_gl_screen_draw_sprite,*/
	NULL,						//AL_METHOD(void, draw_256_sprite, (struct BITMAP *bmp, struct BITMAP *sprite, int x, int y));
	NULL,/*allegro_gl_screen_draw_sprite_v_flip,*/
	NULL,/*allegro_gl_screen_draw_sprite_h_flip,*/
	NULL,/*allegro_gl_screen_draw_sprite_vh_flip,*/
	NULL,						//AL_METHOD(void, draw_trans_sprite, (struct BITMAP *bmp, struct BITMAP *sprite, int x, int y));
	NULL,						//AL_METHOD(void, draw_trans_rgba_sprite, (struct BITMAP *bmp, struct BITMAP *sprite, int x, int y));
	NULL,						//AL_METHOD(void, draw_lit_sprite, (struct BITMAP *bmp, struct BITMAP *sprite, int x, int y, int color));
	NULL,						//AL_METHOD(void, draw_rle_sprite, (struct BITMAP *bmp, struct RLE_SPRITE *sprite, int x, int y));
	NULL,						//AL_METHOD(void, draw_trans_rle_sprite, (struct BITMAP *bmp, struct RLE_SPRITE *sprite, int x, int y));
	NULL,						//AL_METHOD(void, draw_trans_rgba_rle_sprite, (struct BITMAP *bmp, struct RLE_SPRITE *sprite, int x, int y));
	NULL,						//AL_METHOD(void, draw_lit_rle_sprite, (struct BITMAP *bmp, struct RLE_SPRITE *sprite, int x, int y, int color));
	NULL,						//AL_METHOD(void, draw_character, (struct BITMAP *bmp, struct BITMAP *sprite, int x, int y, int color));
	NULL,/*allegro_gl_screen_draw_glyph,*/
	allegro_gl_video_blit_from_memory,
	allegro_gl_video_blit_to_memory,
	NULL,						//AL_METHOD(void, blit_from_system, (struct BITMAP *source, struct BITMAP *dest, int source_x, int source_y, int dest_x, int dest_y, int width, int height));
	NULL,						//AL_METHOD(void, blit_to_system, (struct BITMAP *source, struct BITMAP *dest, int source_x, int source_y, int dest_x, int dest_y, int width, int height));
	allegro_gl_screen_blit_to_self, /* Video bitmaps use same method as screen */
	allegro_gl_screen_blit_to_self, /* ..._forward */
	allegro_gl_screen_blit_to_self, /* ..._backward */
	allegro_gl_memory_blit_between_formats,
	NULL,						//AL_METHOD(void, masked_blit, (struct BITMAP *source, struct BITMAP *dest, int source_x, int source_y, int dest_x, int dest_y, int width, int height));
	allegro_gl_video_clear_to_color,
	NULL,						//AL_METHOD(void, pivot_scaled_sprite_flip, (struct BITMAP *bmp, struct BITMAP *sprite, fixed x, fixed y, fixed cx, fixed cy, fixed angle, fixed scale, int v_flip));
	NULL,/*allegro_gl_screen_draw_sprite_end,*/
	NULL						//AL_METHOD(void, blit_end, (void));
};

